/***************************************************************************\
  "re.h" - Runtime Engine (Phase E)
  Revision 19; 10 Nov 2002 - 21 Feb 2004
  Copyright (C) 2002-2004  Kevin P. Albrecht
\***************************************************************************/

#ifndef RE_H
#define RE_H

/////////////////////////
// DEFINES //////////////
/////////////////////////

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>

// Local //
#include "global.h"
#include "error.h"
#include "icm.h"
#include "k_string.h"
#include "k_conv.h"
#include "rstack.h"

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

// Internal //
typedef enum
{
  fioReading, fioWriting
} FIOMode;

typedef struct FileElement_tag
{
  FILE *file;
  FIOMode mode;
} FileElement;

FileElement *mFiles[GLOBAL_RE_MAX_FILE_HANDLES];
Statement *mSM;
RStack *mExpStack, *mCallStack, *mLocalStack, *mParamStack;
TokenCode mOpCode;

// External //
void re_init ();
void re_deinit ();
void re_run ();

// Internal - Run //
void mRunStatement ();
void mRunJump ();
void mRunCall ();
void mRunReturn ();
void mRunPop ();
void mRunPush ();
void mRunTypeConversion ();
void mRunLogic ();
void mRunUnaryLogic ();
void mRunUnaryArith ();
void mRunIntResultArith ();
void mRunRatResultArith ();
void mRunStrArith ();
void mRunCopy ();
void mRunSwap ();
void mRunASCIIConversion ();
void mRunRandom ();
void mRunConsoleInput ();
void mRunConsoleOutput ();
void mRunFileIO ();

// Internal - Special //
void mGetStatement ();
void mStackPushSymbol (RStack *stack, Symbol *symbol);
void mStackPush (RStack *stack, Type type, void *value);

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

//// re_init
//// Initialize re module
void re_init ()
{
  int i;

  for (i = 0; i < GLOBAL_RE_MAX_FILE_HANDLES; i++)
  {
    mFiles[i] = NULL;
  }

  mExpStack = rstack_new ();
  mCallStack = rstack_new ();
  mLocalStack = rstack_new ();
  mParamStack = rstack_new ();

  srand((unsigned)time(NULL));
}

//// re_deinit
//// Deinitialize re module
void re_deinit ()
{
  int i;

  // Ensure that all files are closed
  for (i = 0; i < GLOBAL_RE_MAX_FILE_HANDLES; i++)
  {
    if (mFiles[i] != NULL)
    {
      fclose (mFiles[i]->file);
      KFREE(mFiles[i]);
      mFiles[i] = NULL;
    }
  }

  rstack_del (mExpStack);
  rstack_del (mCallStack);
  rstack_del (mLocalStack);
  rstack_del (mParamStack);
}

//// re_run
void re_run ()
{
  mGetStatement();
  gnCurrentLine = 0;

  while (mOpCode != ttSTOP)
  {
    mRunStatement ();
    mGetStatement ();
  }
}

/////////
// Run //
/////////

//// mRunStatement
void mRunStatement ()
{
  switch (mOpCode)
  {
    case (ttNOP):
      // do nothing //
      break;
    case (ttDBGON):
      gbDebugErrorsOn = true;
      break;
    case (ttDBGLINE):
      gnDebugLine = *((int*)mSM->sym->value);
      break;
    case (ttDBGFILE):
      k_del (gsDebugFilename);
      gsDebugFilename = k_clone (mSM->sym->value);
      break;
    case (ttJT):
    case (ttJF):
    case (ttJMP):
      mRunJump ();
      break;
    case (ttCALL):
      mRunCall ();
      break;
    case (ttRET):
      mRunReturn ();
      break;
    case (ttPOP):
    case (ttPOPL):
    case (ttPOPP):
      mRunPop ();
      break;
    case (ttPUSH):
    case (ttPUSHL):
    case (ttPUSHP):
      mRunPush ();
      break;
    case (ttTOI):
    case (ttTOS):
    case (ttTOR):
    case (ttTOB):
      mRunTypeConversion ();
      break;
    case (ttEQ):
    case (ttNEQ):
    case (ttLT):
    case (ttGT):
    case (ttLTEQ):
    case (ttGTEQ):
    case (ttOR):
    case (ttAND):
      mRunLogic ();
      break;
    case (ttNOT):
      mRunUnaryLogic ();
      break;
    case (ttNEG):
      mRunUnaryArith ();
      break;
    case (ttADDI):
    case (ttSUBI):
    case (ttMULI):
    case (ttDIVI):
    case (ttPOWI):
    case (ttMODI):
      mRunIntResultArith ();
      break;
    case (ttADDR):
    case (ttSUBR):
    case (ttMULR):
    case (ttDIVR):
    case (ttPOWR):
    case (ttMODR):
      mRunRatResultArith ();
      break;
    case (ttCAT):
      mRunStrArith ();
      break;
    case (ttCOPY):
      mRunCopy ();
      break;
    case (ttSWAP):
      mRunSwap ();
      break;
    case (ttASC):
      mRunASCIIConversion ();
      break;
    case (ttRAND):
      mRunRandom ();
      break;
    case (ttCIN):
      mRunConsoleInput ();
      break;    
    case (ttCOUT):
      mRunConsoleOutput ();
      break;
    case (ttFIO_OR):
    case (ttFIO_OW):
    case (ttFIO_C):
    case (ttFIO_EOF):
    case (ttFIO_R):
    case (ttFIO_W):
      mRunFileIO ();
      break;
    default:
      err (true, ER_INVALID_INSTRUCTION, gvsTokenCodeStrings[mOpCode]);
      erre (EFF_TERMINATE);
      break;
  }
}

//// mRunJump
void mRunJump ()
{
  if (mSM->sym->special != true)
  {
    err (true, ER_LABEL_NOT_DEFINED, mSM->sym->key);
    erre (EFF_TERMINATE);
  }

  switch (mOpCode)
  {
    case (ttJT):
      if (rstack_peektype1 (mExpStack) == tBool)
      {
        void *value = rstack_pop (mExpStack);

        if (*(bool*)value == true)
        {
          icm_goto (*(int*)(mSM->sym->value));
        }

        KFREE(value);
      }
      break;
    case (ttJF):
      if (rstack_peektype1 (mExpStack) == tBool)
      {
        void *value = rstack_pop (mExpStack);

        if (*(bool*)value == false)
        {
          icm_goto (*(int*)(mSM->sym->value));
        }

        KFREE(value);
      }
      break;
    case (ttJMP):
      icm_goto (*(int*)(mSM->sym->value));
      break;
  }
}

//// mRunCall
void mRunCall ()
{
  if (mSM->sym->special != true)
  {
    err (true, ER_ROUTINE_LABEL_NOT_DEFINED, mSM->sym->key);
    erre (EFF_TERMINATE);
  }

  // push return address onto the call stack //
  mStackPush (mCallStack, tLabel, &gxCurr);

  // jump to the first instruction of the routine //
  icm_goto ( *(int *)(mSM->sym->value) );
}

//// mRunReturn
void mRunReturn ()
{
  void *return_value;

  return_value = rstack_pop (mCallStack);
    
  // jump to the next instruction of the calling routine //
  icm_goto (*(int*)(return_value));
  
  KFREE(return_value);
}

void mRunPop ()
{
  RStack *stack;

  switch (mOpCode)
  {
    case (ttPOP):
      stack = mExpStack;
      break;
    case (ttPOPL):
      stack = mLocalStack;
      break;
    case (ttPOPP):
      stack = mParamStack;
      break;
  }

  // types are the same //
  if (mSM->sym->type == rstack_peektype1 (stack))
  {
    KFREE(mSM->sym->value);

    mSM->sym->value = rstack_pop (stack);

    if ((mSM->sym->value == NULL) && (mSM->sym->type != tStr)) // only error if not str, strings can be NULL
    {
      err (true, ER_STACK_EMPTY, NULL);
      erre (EFF_TERMINATE);
    }
  }
  // types are different //
  else
  {
    err (true, ER_POP_MISMATCHED_TYPE, gvsTypeStrings[rstack_peektype1(stack)]);
    erre (EFF_TERMINATE);
  }
}

void mRunPush ()
{
  RStack *stack;

  switch (mOpCode)
  {
    case (ttPUSH):
      stack = mExpStack;
      break;
    case (ttPUSHL):
      stack = mLocalStack;
      break;
    case (ttPUSHP):
      stack = mParamStack;
      break;
  }

  mStackPushSymbol (stack, mSM->sym);
}

//// mRunConsoleOutput
void mRunConsoleOutput ()
{
  Type type = rstack_peektype1 (mExpStack);
  void *value = rstack_pop (mExpStack);

  switch (type)
  {
    case (tBool):
    {
      bool bValue = *(bool *)value;

      if (bValue == true) printf ("true");
      else                printf ("false");
      break;
    }
    case (tInt):
      printf ("%d", *(int *)value);
      break;
    case (tRat):
      printf ("%.16f", *(double *)value);
      break;
    case (tStr):
      printf ("%s", (char *)value);
      break;
  }

  KFREE(value);
}

void mRunTypeConversion ()
{
  bool btemp;
  int ntemp;
  double dtemp;
  char *stemp;
  Type type;
  void *value;

  btemp = false;
  ntemp = 0;
  dtemp = 0.0;
  stemp = NULL;
  type = rstack_peektype1 (mExpStack);
  value = rstack_pop (mExpStack);

  switch (mOpCode)
  {
    // destination type is bool //
    case (ttTOB):
    {
      switch (type)
      {
        case tBool:
          btemp = *(bool*) value;
          break;
        case tInt:
          btemp = (bool) *(int*) value;
          break;
        case tRat:
          btemp = (bool) *(double*) value;
          break;
        case tStr:
          stemp = (char*) value;
          btemp = k_conv_sb (stemp);
          break;
      }
      mStackPush (mExpStack, tBool, &btemp);          
      break;
    }

    // destination type is int //
    case (ttTOI):
    {
      switch (type)
      {
        case tBool:
          btemp = *(bool*) value;
          if (btemp) ntemp = 1;
          else       ntemp = 0;
          break;
        case tInt:
          ntemp = *(int*) value;
          break;
        case tRat:
          ntemp = (int) *(double*) value;
          break;
        case tStr:
          ntemp = k_conv_si((char*) value);
          break;
      }
      mStackPush (mExpStack, tInt, &ntemp);          
      break;
    }

    // destination type is rat //
    case (ttTOR):
    {
      switch (type)
      {
        case tBool:
          btemp = *(bool*) value;
          if (btemp) dtemp = 1.0;
          else       dtemp = 0.0;
          break;
        case tInt:
          dtemp = (double) *(int*) value;
          break;
        case tRat:
          dtemp = *(double*) value;
          break;
        case tStr:
          dtemp = k_conv_sd((char*) value);
          break;
      }
      mStackPush (mExpStack, tRat, &dtemp);          
      break;
    }

    // destination type is str //
    case (ttTOS):
    {
      switch (type)
      {
        case tBool:
          stemp = k_conv_bs(*(bool*) value);
          break;
        case tInt:
          stemp = k_conv_is(*(int*) value);
          break;
        case tRat:
          stemp = k_conv_ds(*(double*) value);
          break;
        case tStr:
          stemp = k_clone((char*) value);
          break;
      }
      mStackPush (mExpStack, tStr, stemp);          
      break;
    }
  }

  KFREE (value);
}

//// mRunASCIIConversion
void mRunASCIIConversion ()
{
  char source;
  char *dest;
  Type type;
  void* value;

  type = rstack_peektype1 (mExpStack);
  value = rstack_pop (mExpStack);

  if (type != tInt)
  {
    err (true, ER_POP_MISMATCHED_TYPE, NULL);
    errm ("Expected int, got",  gvsTypeStrings[type]);
    erre (EFF_TERMINATE);
  }

  source = (char) *(int *)value;

  if (source == 0)
  {
    err (true, ER_ASCII_ZERO, NULL);
    erre (EFF_TERMINATE);
  }

  dest = k_new (2);
  dest[0] = source;
  dest[1] = '\0';

  mStackPush (mExpStack, tStr, dest);
  k_del (dest);
  KFREE (value);
}

//// mRunRandom
void mRunRandom ()
{
  int num;

  num = rand();

  mStackPush (mExpStack, tInt, &num);
}

void mRunCopy ()
{
  rstack_copytop (mExpStack);
}

void mRunSwap ()
{
  rstack_swap (mExpStack);
}

//// mRunConsoleInput
void mRunConsoleInput ()
{
  char input[GLOBAL_RE_CIN_LINE_LEN];

  // Loop until the user inputs a non-blank line //
  gets (input);
  while (strlen(input) == 0) gets (input);

  mStackPush (mExpStack, tStr, input); 
}

void mRunLogic ()
{
  Type typea, typeb;
  void *vala;
  void *valb;
  bool result;

  typeb = rstack_peektype1 (mExpStack);
  valb = rstack_pop (mExpStack);
  typea = rstack_peektype1 (mExpStack);
  vala = rstack_pop (mExpStack);

  switch (mOpCode)
  {
    case (ttEQ):
    {
      if ( (typea == tStr) && (typeb == tStr) )
      {
        if (k_same ((char*)vala, (char*)valb))
          result = true;
        else
          result = false;
      }
      else if ( (typea == tInt) && (typeb == tInt) )
      {
        result = ( *(int*)vala == *(int*)valb );
      }
      else if ( (typea == tInt) && (typeb == tRat) )
      {
        result = ( *(int*)vala == *(double*)valb );
      }
      else if ( (typea == tRat) && (typeb == tInt) )
      {
        result = ( *(double*)vala == *(int*)valb );
      }
      else if ( (typea == tRat) && (typeb == tRat) )
      {
        result = ( *(double*)vala == *(double*)valb );
      }
      else if ( (typea == tBool) && (typeb == tBool) )
      {
        result = ( *(bool*)vala == *(bool*)valb );
      }
      else
      {
        err (true, ER_MISMATCHED_TYPES, NULL);
        erre (EFF_TERMINATE);
      }
      break;
    }
    case (ttNEQ):
    {
      if ( (typea == tStr) && (typeb == tStr) )
      {
        if (k_same ((char*)vala, (char*)valb))
          result = false;
        else
          result = true;
      }
      else if ( (typea == tInt) && (typeb == tInt) )
      {
        result = ( *(int*)vala != *(int*)valb );
      }
      else if ( (typea == tInt) && (typeb == tRat) )
      {
        result = ( *(int*)vala != *(double*)valb );
      }
      else if ( (typea == tRat) && (typeb == tInt) )
      {
        result = ( *(double*)vala != *(int*)valb );
      }
      else if ( (typea == tRat) && (typeb == tRat) )
      {
        result = ( *(double*)vala != *(double*)valb );
      }
      else if ( (typea == tBool) && (typeb == tBool) )
      {
        result = ( *(bool*)vala != *(bool*)valb );
      }
      else
      {
        err (true, ER_MISMATCHED_TYPES, NULL);
        erre (EFF_TERMINATE);
      }
      break;
    }
    case (ttLT):
    {
      if ( (typea == tInt) && (typeb == tInt) )
      {
        result = ( *(int*)vala < *(int*)valb );
      }
      else if ( (typea == tInt) && (typeb == tRat) )
      {
        result = ( *(int*)vala < *(double*)valb );
      }
      else if ( (typea == tRat) && (typeb == tInt) )
      {
        result = ( *(double*)vala < *(int*)valb );
      }
      else if ( (typea == tRat) && (typeb == tRat) )
      {
        result = ( *(double*)vala < *(double*)valb );
      }
      else
      {
        err (true, ER_MISMATCHED_TYPES, NULL);
        erre (EFF_TERMINATE);
      }
      break;
    }
    case (ttGT):
    {
      if ( (typea == tInt) && (typeb == tInt) )
      {
        result = ( *(int*)vala > *(int*)valb );
      }
      else if ( (typea == tInt) && (typeb == tRat) )
      {
        result = ( *(int*)vala > *(double*)valb );
      }
      else if ( (typea == tRat) && (typeb == tInt) )
      {
        result = ( *(double*)vala > *(int*)valb );
      }
      else if ( (typea == tRat) && (typeb == tRat) )
      {
        result = ( *(double*)vala > *(double*)valb );
      }
      else
      {
        err (true, ER_MISMATCHED_TYPES, NULL);
        erre (EFF_TERMINATE);
      }
      break;
    }
    case (ttLTEQ):
    {
      if ( (typea == tInt) && (typeb == tInt) )
      {
        result = ( *(int*)vala <= *(int*)valb );
      }
      else if ( (typea == tInt) && (typeb == tRat) )
      {
        result = ( *(int*)vala <= *(double*)valb );
      }
      else if ( (typea == tRat) && (typeb == tInt) )
      {
        result = ( *(double*)vala <= *(int*)valb );
      }
      else if ( (typea == tRat) && (typeb == tRat) )
      {
        result = ( *(double*)vala <= *(double*)valb );
      }
      else
      {
        err (true, ER_MISMATCHED_TYPES, NULL);
        erre (EFF_TERMINATE);
      }
      break;
    }
    case (ttGTEQ):
    {
      if ( (typea == tInt) && (typeb == tInt) )
      {
        result = ( *(int*)vala >= *(int*)valb );
      }
      else if ( (typea == tInt) && (typeb == tRat) )
      {
        result = ( *(int*)vala >= *(double*)valb );
      }
      else if ( (typea == tRat) && (typeb == tInt) )
      {
        result = ( *(double*)vala >= *(int*)valb );
      }
      else if ( (typea == tRat) && (typeb == tRat) )
      {
        result = ( *(double*)vala >= *(double*)valb );
      }
      else
      {
        err (true, ER_MISMATCHED_TYPES, NULL);
        erre (EFF_TERMINATE);
      }
      break;
    }
    case (ttOR):
    {
      if ( (typea == tBool) && (typeb == tBool) )
      {
        result = ( *(bool*)vala || *(bool*)valb );
      }
      else
      {
        err (true, ER_MISMATCHED_TYPES, NULL);
        erre (EFF_TERMINATE);
      }
      break;
    }
    case (ttAND):
    {
      if ( (typea == tBool) && (typeb == tBool) )
      {
        result = ( *(bool*)vala && *(bool*)valb );
      }
      else
      {
        err (true, ER_MISMATCHED_TYPES, NULL);
        erre (EFF_TERMINATE);
      }
      break;
    }
  }

  mStackPush (mExpStack, tBool, &result);
  KFREE(vala);
  KFREE(valb);
}

void mRunUnaryLogic ()
{
  Type type;
  void *val;
  bool result;

  type = rstack_peektype1 (mExpStack);
  val = rstack_pop (mExpStack);

  switch (mOpCode)
  {
    case (ttNOT):
    {
      if (type == tBool)
      {
        result = ! (*(bool*) val);
      }
      else
      {
        err (true, ER_MISMATCHED_TYPES, NULL);
        erre (EFF_TERMINATE);
      }
      break;
    }
  }

  mStackPush (mExpStack, tBool, &result);
  KFREE(val);
}

void mRunUnaryArith ()
{
  Type type;
  void *val;

  type = rstack_peektype1 (mExpStack);
  val = rstack_pop (mExpStack);

  switch (mOpCode)
  {
    case (ttNEG):
    {
      if (type == tInt)
      {
        int result = - (*(int*) val);
        mStackPush (mExpStack, tInt, &result);
      }
      else if (type == tRat)
      {
        double result = - (*(double*) val);
        mStackPush (mExpStack, tRat, &result);
      }
      else
      {
        err (true, ER_MISMATCHED_TYPES, NULL);
        erre (EFF_TERMINATE);
      }
      break;
    }
  }

  KFREE(val);
}

void mRunStrArith ()
{
  Type typea, typeb;
  void *vala;
  void *valb;
  char *temp;

  typeb = rstack_peektype1 (mExpStack);
  valb = rstack_pop (mExpStack);
  typea = rstack_peektype1 (mExpStack);
  vala = rstack_pop (mExpStack);

  if ( (typea != tStr) || (typeb != tStr) )
  {
    err (true, ER_MISMATCHED_TYPES, NULL);
    erre (EFF_TERMINATE);
  }

  temp = k_new (strlen(vala) + strlen(valb) + 1);
  k_copy (temp, vala);
  k_cat (temp, valb);

  mStackPush (mExpStack, tStr, temp);
  k_del (vala);
  k_del (valb);
  k_del (temp);
}

void mRunIntResultArith ()
{
  Type typea, typeb;
  void *vala;
  void *valb;
  int result;

  typeb = rstack_peektype1 (mExpStack);
  valb = rstack_pop (mExpStack);
  typea = rstack_peektype1 (mExpStack);
  vala = rstack_pop (mExpStack);

  switch (mOpCode)
  {
    case (ttADDI):
    {
      if ( (typea == tInt) && (typeb == tInt) )
      {
        result = (int) ( *(int*)vala + *(int*)valb );
      }
      else if ( (typea == tInt) && (typeb == tRat) )
      {
        result = (int) ( *(int*)vala + *(double*)valb );
      }
      else if ( (typea == tRat) && (typeb == tInt) )
      {
        result = (int) ( *(double*)vala + *(int*)valb );
      }
      else if ( (typea == tRat) && (typeb == tRat) )
      {
        result = (int) ( *(double*)vala + *(double*)valb );
      }
      else
      {
        err (true, ER_MISMATCHED_TYPES, NULL);
        erre (EFF_TERMINATE);
      }
      break;
    }
    case (ttSUBI):
    {
      if ( (typea == tInt) && (typeb == tInt) )
      {
        result = (int) ( *(int*)vala - *(int*)valb );
      }
      else if ( (typea == tInt) && (typeb == tRat) )
      {
        result = (int) ( *(int*)vala - *(double*)valb );
      }
      else if ( (typea == tRat) && (typeb == tInt) )
      {
        result = (int) ( *(double*)vala - *(int*)valb );
      }
      else if ( (typea == tRat) && (typeb == tRat) )
      {
        result = (int) ( *(double*)vala - *(double*)valb );
      }
      else
      {
        err (true, ER_MISMATCHED_TYPES, NULL);
        erre (EFF_TERMINATE);
      }
      break;
    }
    case (ttMULI):
    {
      if ( (typea == tInt) && (typeb == tInt) )
      {
        result = (int) ( *(int*)vala * *(int*)valb );
      }
      else if ( (typea == tInt) && (typeb == tRat) )
      {
        result = (int) ( *(int*)vala * *(double*)valb );
      }
      else if ( (typea == tRat) && (typeb == tInt) )
      {
        result = (int) ( *(double*)vala * *(int*)valb );
      }
      else if ( (typea == tRat) && (typeb == tRat) )
      {
        result = (int) ( *(double*)vala * *(double*)valb );
      }
      else
      {
        err (true, ER_MISMATCHED_TYPES, NULL);
        erre (EFF_TERMINATE);
      }
      break;
    }
    case (ttDIVI):
    {
      if ( (typea == tInt) && (typeb == tInt) )
      {
        if (*(int*)valb == 0)
        {
          err (true, ER_DIV_BY_ZERO, NULL);
          erre (EFF_TERMINATE);
        }

        result = (int) ( *(int*)vala / *(int*)valb );
      }
      else if ( (typea == tInt) && (typeb == tRat) )
      {
        if (*(double*)valb == 0)
        {
          err (true, ER_DIV_BY_ZERO, NULL);
          erre (EFF_TERMINATE);
        }

        result = (int) ( *(int*)vala / *(double*)valb );
      }
      else if ( (typea == tRat) && (typeb == tInt) )
      {
        if (*(int*)valb == 0)
        {
          err (true, ER_DIV_BY_ZERO, NULL);
          erre (EFF_TERMINATE);
        }

        result = (int) ( *(double*)vala / *(int*)valb );
      }
      else if ( (typea == tRat) && (typeb == tRat) )
      {
        if (*(double*)valb == 0)
        {
          err (true, ER_DIV_BY_ZERO, NULL);
          erre (EFF_TERMINATE);
        }

        result = (int) ( *(double*)vala / *(double*)valb );
      }
      else
      {
        err (true, ER_MISMATCHED_TYPES, NULL);
        erre (EFF_TERMINATE);
      }
      break;
    }
    case (ttPOWI):
    {
      if ( (typea == tInt) && (typeb == tInt) )
      {
        result = (int) floor (pow ( (double) *(int*)vala , (double) *(int*)valb ));
      }
      else if ( (typea == tInt) && (typeb == tRat) )
      {
        result = (int) floor (pow ( (double) *(int*)vala , *(double*)valb ));
      }
      else if ( (typea == tRat) && (typeb == tInt) )
      {
        result = (int) floor (pow ( *(double*)vala , (double) *(int*)valb ));
      }
      else if ( (typea == tRat) && (typeb == tRat) )
      {
        result = (int) floor (pow ( *(double*)vala , *(double*)valb ));
      }
      else
      {
        err (true, ER_MISMATCHED_TYPES, NULL);
        erre (EFF_TERMINATE);
      }
      break;
    }
    case (ttMODI):
    {
      if ( (typea == tInt) && (typeb == tInt) )
      {
        if (*(int*)valb == 0)
        {
          err (true, ER_DIV_BY_ZERO, NULL);
          erre (EFF_TERMINATE);
        }

        result = (int) ( *(int*)vala % *(int*)valb );
      }
      else
      {
        err (true, ER_MISMATCHED_TYPES, NULL);
        erre (EFF_TERMINATE);
      }
      break;
    }
  }

  mStackPush (mExpStack, tInt, &result);
  KFREE(vala);
  KFREE(valb);
}

void mRunRatResultArith ()
{
  Type typea, typeb;
  void *vala;
  void *valb;
  double result;

  typeb = rstack_peektype1 (mExpStack);
  valb = rstack_pop (mExpStack);
  typea = rstack_peektype1 (mExpStack);
  vala = rstack_pop (mExpStack);

  switch (mOpCode)
  {
    case (ttADDR):
    {
      if ( (typea == tInt) && (typeb == tInt) )
      {
        result = (double) ( *(int*)vala + *(int*)valb );
      }
      else if ( (typea == tInt) && (typeb == tRat) )
      {
        result = (double) ( *(int*)vala + *(double*)valb );
      }
      else if ( (typea == tRat) && (typeb == tInt) )
      {
        result = (double) ( *(double*)vala + *(int*)valb );
      }
      else if ( (typea == tRat) && (typeb == tRat) )
      {
        result = (double) ( *(double*)vala + *(double*)valb );
      }
      else
      {
        err (true, ER_MISMATCHED_TYPES, NULL);
        erre (EFF_TERMINATE);
      }
      break;
    }
    case (ttSUBR):
    {
      if ( (typea == tInt) && (typeb == tInt) )
      {
        result = (double) ( *(int*)vala - *(int*)valb );
      }
      else if ( (typea == tInt) && (typeb == tRat) )
      {
        result = (double) ( *(int*)vala - *(double*)valb );
      }
      else if ( (typea == tRat) && (typeb == tInt) )
      {
        result = (double) ( *(double*)vala - *(int*)valb );
      }
      else if ( (typea == tRat) && (typeb == tRat) )
      {
        result = (double) ( *(double*)vala - *(double*)valb );
      }
      else
      {
        err (true, ER_MISMATCHED_TYPES, NULL);
        erre (EFF_TERMINATE);
      }
      break;
    }
    case (ttMULR):
    {
      if ( (typea == tInt) && (typeb == tInt) )
      {
        result = (double) ( *(int*)vala * *(int*)valb );
      }
      else if ( (typea == tInt) && (typeb == tRat) )
      {
        result = (double) ( *(int*)vala * *(double*)valb );
      }
      else if ( (typea == tRat) && (typeb == tInt) )
      {
        result = (double) ( *(double*)vala * *(int*)valb );
      }
      else if ( (typea == tRat) && (typeb == tRat) )
      {
        result = (double) ( *(double*)vala * *(double*)valb );
      }
      else
      {
        err (true, ER_MISMATCHED_TYPES, NULL);
        erre (EFF_TERMINATE);
      }
      break;
    }
    case (ttDIVR):
    {
      if ( (typea == tInt) && (typeb == tInt) )
      {
        if (*(int*)valb == 0)
        {
          err (true, ER_DIV_BY_ZERO, NULL);
          erre (EFF_TERMINATE);
        }

        result = (double) ( *(int*)vala / *(int*)valb );
      }
      else if ( (typea == tInt) && (typeb == tRat) )
      {
        if (*(double*)valb == 0)
        {
          err (true, ER_DIV_BY_ZERO, NULL);
          erre (EFF_TERMINATE);
        }

        result = (double) ( *(int*)vala / *(double*)valb );
      }
      else if ( (typea == tRat) && (typeb == tInt) )
      {
        if (*(int*)valb == 0)
        {
          err (true, ER_DIV_BY_ZERO, NULL);
          erre (EFF_TERMINATE);
        }

        result = (double) ( *(double*)vala / *(int*)valb );
      }
      else if ( (typea == tRat) && (typeb == tRat) )
      {
        if (*(double*)valb == 0)
        {
          err (true, ER_DIV_BY_ZERO, NULL);
          erre (EFF_TERMINATE);
        }

        result = (double) ( *(double*)vala / *(double*)valb );
      }
      else
      {
        err (true, ER_MISMATCHED_TYPES, NULL);
        erre (EFF_TERMINATE);
      }
      break;
    }
    case (ttPOWR):
    {
      if ( (typea == tInt) && (typeb == tInt) )
      {
        result = (double) pow ( (double) *(int*)vala , (double) *(int*)valb );
      }
      else if ( (typea == tInt) && (typeb == tRat) )
      {
        result = (double) pow ( (double) *(int*)vala , *(double*)valb );
      }
      else if ( (typea == tRat) && (typeb == tInt) )
      {
        result = (double) pow ( *(double*)vala , (double) *(int*)valb );
      }
      else if ( (typea == tRat) && (typeb == tRat) )
      {
        result = (double) pow ( *(double*)vala , *(double*)valb );
      }
      else
      {
        err (true, ER_MISMATCHED_TYPES, NULL);
        erre (EFF_TERMINATE);
      }
      break;
    }
    case (ttMODR):
    {
      if ( (typea == tInt) && (typeb == tInt) )
      {
        if (*(int*)valb == 0)
        {
          err (true, ER_DIV_BY_ZERO, NULL);
          erre (EFF_TERMINATE);
        }

        result = (double) ( *(int*)vala % *(int*)valb );
      }
      else
      {
        err (true, ER_MISMATCHED_TYPES, NULL);
        erre (EFF_TERMINATE);
      }
      break;
    }
  }

  mStackPush (mExpStack, tRat, &result);
  KFREE(vala);
  KFREE(valb);
}

/*
fio_or x:str    ( -- y:int )
fio_ow x:str    ( -- y:int )
  x = src filename  
  y = dest handle

fio_c x:int     ( -- )
  x = src handle

fio_eof x:int ( -- y:bool )
  x = src handle
  y = "true" if end of file reached, "false otherwise"

fio_r x:int     ( -- y:str )
  x = src handle
  y = dest of input

fio_w x:int     ( y:str -- )
  x = src handle
  y = src of output
*/
  
//// mRunFileIO
void mRunFileIO ()
{
  int i;
  void *value;
  char *sGood;
  char tempbuf[GLOBAL_RE_FIO_LINE_LEN];

  switch (mOpCode)
  {
    case (ttFIO_OR): // open file for read
    {
      for (i = 0; i < GLOBAL_RE_MAX_FILE_HANDLES; i++)
      {
        if (mFiles[i] == NULL) break;
      }
      if (i == GLOBAL_RE_MAX_FILE_HANDLES)
      {
        err (true, ER_MAX_FILES_OPENED, NULL);
        erre (EFF_TERMINATE);
      }

      value = k_new_n (i);
      mStackPush (mExpStack, tInt, value);
      KFREE(value);

      mFiles[i] = (FileElement *) KALLOC(sizeof (FileElement));
      mFiles[i]->file = fopen ((char *)mSM->sym->value, "r");
      mFiles[i]->mode = fioReading;

      if (!(mFiles[i]->file))
      {
        err (true, ER_FILE_OPEN_ERROR, (char *)(mSM->sym->value));
        erre (EFF_TERMINATE);
      }

      break;
    }
    case (ttFIO_OW): // open file for write
    {
      for (i = 0; i < GLOBAL_RE_MAX_FILE_HANDLES; i++)
      {
        if (mFiles[i] == NULL) break;
      }
      if (i == GLOBAL_RE_MAX_FILE_HANDLES)
      {
        err (true, ER_MAX_FILES_OPENED, NULL);
        erre (EFF_TERMINATE);
      }

      value = k_new_n (i);
      mStackPush (mExpStack, tInt, value);
      KFREE(value);

      mFiles[i] = (FileElement *) KALLOC(sizeof (FileElement));
      mFiles[i]->file = fopen ((char *)mSM->sym->value, "w");
      mFiles[i]->mode = fioWriting;

      if (!(mFiles[i]->file))
      {
        err (true, ER_FILE_OPEN_ERROR, (char *)(mSM->sym->value));
        erre (EFF_TERMINATE);
      }

      break;
    }
    case (ttFIO_C): // close
    {
      i = *(int *)mSM->sym->value;
      if (mFiles[i] != NULL)
      {
        fclose (mFiles[i]->file);
        KFREE(mFiles[i]);
        mFiles[i] = NULL;
      }
      else
      {
        err (true, ER_INVALID_FILE_HANDLE, NULL);
        erre (EFF_IGNORE);
      }
      break;
    }
    case (ttFIO_EOF): // end-of-file check
    {
      i = *(int *)mSM->sym->value;
      if (mFiles[i] != NULL)
      {
        if (mFiles[i]->mode == fioReading)
        {
          bool bValue = false;
          
          if (feof (mFiles[i]->file))
          {
            bValue = true;
          }

          mStackPush (mExpStack, tBool, &bValue);
        }
        else
        {
          err (true, ER_INVALID_FILE_MODE, NULL);
          erre (EFF_IGNORE);
        }
      }
      else
      {
        err (true, ER_INVALID_FILE_HANDLE, NULL);
        erre (EFF_IGNORE);
      }
      break;
    }
    case (ttFIO_R): // read
    {
      i = *(int *)mSM->sym->value;
      if (mFiles[i] != NULL)
      {
        if (mFiles[i]->mode == fioReading)
        {
          // Check to make sure EOF has not been reached //
          if (!feof (mFiles[i]->file))
          {
            sGood = fgets (tempbuf, GLOBAL_RE_FIO_LINE_LEN, mFiles[i]->file);
            if (sGood)
            {
              value = k_clone (tempbuf);
              mStackPush (mExpStack, tStr, value);
              KFREE(value);
            }
            else
            {
              value = k_clone ("");
              mStackPush (mExpStack, tStr, value);
              KFREE(value);
            }
          }

          // EOF was reached, cause error //
          else
          {
            err (true, ER_FIO_R_EOF_REACHED, NULL);
            erre (EFF_TERMINATE);
          }
        }
        else
        {
          err (true, ER_INVALID_FILE_MODE, NULL);
          erre (EFF_IGNORE);
        }
      }
      else
      {
        err (true, ER_INVALID_FILE_HANDLE, NULL);
        erre (EFF_IGNORE);
      }
      break;
    }
    case (ttFIO_W): // write
    {
      i = *(int *)mSM->sym->value;

      if (rstack_peektype1 (mExpStack) != tStr)
      {
        err (true, ER_MISMATCHED_TYPES, NULL);
        erre (EFF_TERMINATE);
      }

      value = rstack_pop (mExpStack);

      if (mFiles[i] != NULL)
      {
        if (mFiles[i]->mode == fioWriting)
        {
          fprintf (mFiles[i]->file, "%s", (char*)value);
          KFREE(value);
        }
        else
        {
          err (true, ER_INVALID_FILE_MODE, NULL);
          erre (EFF_IGNORE);
        }
      }
      else
      {
        err (true, ER_INVALID_FILE_HANDLE, NULL);
        erre (EFF_IGNORE);
      }
      break;
    }
  }
}

//////////////
// Special  //
//////////////

void mGetStatement ()
{
  mSM = icm_getnext();
  if (mSM)
  {
    mOpCode = mSM->code;
    k_copy (gsCurrentFilename, mSM->file);
    gnCurrentLine = mSM->line;
  }
  else
  {
    err (true, ER_END_OF_FILE, NULL);
    errm ("Program must be terminated with a \"stop\" instruction", NULL);
    erre (EFF_TERMINATE);
  }
}

void mStackPush (RStack *stack, Type type, void *value)
{
  Symbol *symbol;

  // push return address onto the runtime stack //
  symbol = sym_new ("pushed value");
  sym_set (symbol, repLiteral, type, value);

  rstack_push (stack, symbol);

  sym_del (symbol);
}

void mStackPushSymbol (RStack *stack, Symbol *symbol)
{
  rstack_push (stack, symbol);
}

#endif

